r"""Quantum gates required for Weyl chamber calculations

**Module data:**

.. data:: Qmagic

    Transformation matrix to the "magic" Bell basis:

    .. math::

        \Op{Q} = \frac{1}{\sqrt{2}} \begin{pmatrix}
           1 & 0 &  0 &  i \\
           0 & i &  1 &  0 \\
           0 & i & -1 &  0 \\
           1 & 0 &  0 & -i
        \end{pmatrix}

    See "Theorem 1" in Y. Makhlin, Quantum Inf. Process. 1, 243 (2002)


.. data:: SxSx

    :math:`\Op{\sigma}_x \otimes \Op{\sigma}_x` gate


.. data:: SySy

    :math:`\Op{\sigma}_y \otimes \Op{\sigma}_y` gate


.. data:: SzSz

    :math:`\Op{\sigma}_z \otimes \Op{\sigma}_z` gate
"""
import qutip
import numpy as np

__all__ = [
    'Qmagic',
    'SxSx',
    'SySy',
    'SzSz',
    'bell_basis',
    'gate',
    'mapped_basis',
]


Qmagic = (1.0 / np.sqrt(2.0)) * qutip.Qobj(
    np.array(
        [[1,  0,  0,  1j],
         [0, 1j,  1,  0],
         [0, 1j, -1,  0],
         [1,  0,  0, -1j]],
        dtype=np.complex128,
    ),
    dims=[[2, 2], [2, 2]],
)


SxSx = qutip.Qobj(
    np.array(
        [[0, 0, 0, 1], [0, 0, 1, 0], [0, 1, 0, 0], [1, 0, 0, 0]],
        dtype=np.complex128,
    ),
    dims=[[2, 2], [2, 2]],
)


SySy = qutip.Qobj(
    np.array(
        [[+0, 0, 0, -1], [+0, 0, 1, 0], [+0, 1, 0, 0], [-1, 0, 0, 0]],
        dtype=np.complex128,
    ),
    dims=[[2, 2], [2, 2]],
)


SzSz = qutip.Qobj(
    np.array(
        [[1, 0, 0, 0], [0, -1, 0, 0], [0, 0, -1, 0], [0, 0, 0, 1]],
        dtype=np.complex128,
    ),
    dims=[[2, 2], [2, 2]],
)


def bell_basis(canonical_basis):
    """Two-qubit Bell basis associated with the given canonical basis

    Example:

        >>> from qutip import ket
        >>> canonical_basis = [
        ...     ket(nums) for nums in [(0, 0), (0, 1), (1, 0), (1, 1)]
        ... ]
        >>> bell = bell_basis(canonical_basis)
        >>> _bell = [
        ...     (     ket((0, 0)) +      ket((1, 1))) / np.sqrt(2),
        ...     (1j * ket((0, 1)) + 1j * ket((1, 0))) / np.sqrt(2),
        ...     (     ket((0, 1)) -      ket((1, 0))) / np.sqrt(2),
        ...     (1j * ket((0, 0)) - 1j * ket((1, 1))) / np.sqrt(2),
        ... ]
        >>> assert (bell[0] - _bell[0]).norm() < 1e-15
        >>> assert (bell[1] - _bell[1]).norm() < 1e-15
        >>> for (a, b) in zip(bell, _bell):
        ...     assert (a - b).norm() < 1e-15, (a - b).norm()
    """
    return mapped_basis(Qmagic, canonical_basis)


def gate(basis, states):
    """Two-qubit gate that maps `basis` to `states`

    Example:

        >>> from qutip import ket
        >>> basis = [ket(nums) for nums in [(0, 0), (0, 1), (1, 0), (1, 1)]]
        >>> states = mapped_basis(qutip.gates.cnot(), basis)
        >>> U = gate(basis, states)
        >>> assert (U - qutip.gates.cnot()).norm() < 1e-15
    """
    if not (len(basis) == len(states) == 4):
        raise ValueError(
            "Both `basis` and `states` must be a list of four states"
        )
    U = np.zeros((4, 4), dtype=np.complex128)
    for j in range(4):
        for i in range(4):
            U[i, j] = basis[i].overlap(states[j])
    return qutip.Qobj(U, dims=[[2, 2], [2, 2]])


def mapped_basis(gate, basis):
    """Result of applying `gate` to `basis`

    Example:

        >>> from qutip import ket
        >>> basis = [ket(nums) for nums in [(0, 0), (0, 1), (1, 0), (1, 1)]]
        >>> states = mapped_basis(qutip.gates.cnot(), basis)
        >>> assert (states[0] - ket((0,0))).norm() < 1e-15
        >>> assert (states[1] - ket((0,1))).norm() < 1e-15
        >>> assert (states[2] - ket((1,1))).norm() < 1e-15  # swap (1, 1) ...
        >>> assert (states[3] - ket((1,0))).norm() < 1e-15  # ... and (1, 0)
    """
    return tuple(
        [
            sum([gate[i, j] * basis[i] for i in range(gate.shape[0])])
            for j in range(gate.shape[1])
        ]
    )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import os
import sys
import datetime
import git

import weylchamber

#from unittest import mock
#MOCK_MODULES = ['numpy', 'numpy.linalg', 'scipy', 'scipy.sparse', 'matplotlib',
    #'matplotlib.pyplot']
#sys.modules.update((mod_name, mock.Mock()) for mod_name in MOCK_MODULES)

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.abspath('_extensions'))

# -- Generate API documentation ------------------------------------------------
def run_apidoc(_):
    """Generage API documentation"""
    import better_apidoc
    better_apidoc.main([
        'better-apidoc',
        '-t',
        os.path.join('.', '_templates'),
        '--force',
        '--no-toc',
        '--separate',
        '-o',
        os.path.join('.', 'API'),
        os.path.join('..', 'src', 'weylchamber'),
    ])


# -- General configuration -----------------------------------------------------

# Report broken links as warnings
nitpicky = True
nitpick_ignore = [
    ('py:class', 'callable'),
]

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.coverage',
    'sphinx.ext.mathjax',
    'sphinx.ext.viewcode',
    'sphinx.ext.napoleon',
    'sphinx.ext.intersphinx',
    'sphinx.ext.autosummary',
    'sphinx.ext.extlinks',
    'sphinx.ext.ifconfig',
    'sphinx.ext.todo',
    'sphinx.ext.inheritance_diagram',
    'dollarmath',
    'sphinx_autodoc_typehints',
]

extensions.append('nbsphinx')


if os.getenv('SPELLCHECK'):
    extensions += 'sphinxcontrib.spelling',
    spelling_show_suggestions = True
    spelling_lang = os.getenv('SPELLCHECK')
    spelling_word_list_filename = 'spelling_wordlist.txt'
    spelling_ignore_pypi_package_names = True

intersphinx_mapping = {
    'python': ('https://docs.python.org/3.6', None),
    'sympy': ('https://docs.sympy.org/latest/', None),
    'scipy': ('https://docs.scipy.org/doc/scipy/reference/', None),
    'numpy': ('https://docs.scipy.org/doc/numpy/', None),
    'matplotlib': ('https://matplotlib.org/', None),
    'qutip': ('http://qutip.org/docs/latest/', None),
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

source_suffix = '.rst'
master_doc = 'index'
project = 'weylchamber'
year = str(datetime.datetime.now().year)
author = 'Michael Goerz'
copyright = '{0}, {1}'.format(year, author)
version = weylchamber.__version__
rootfolder = os.path.join(os.path.dirname(__file__), '..')
try:
    last_commit = str(git.Repo(rootfolder).head.commit)[:7]
    release = last_commit
except (git.exc.InvalidGitRepositoryError, ValueError):
    release = version
numfig = True

pygments_style = 'friendly'
extlinks = {
    'issue': ('https://github.com/qucontrol/weylchamber/issues/%s', '#'),
    'pr': ('https://github.com/qucontrol/weylchamber/pull/%s', 'PR #'),
}

# autodoc settings
autoclass_content = 'both'
autodoc_member_order = 'bysource'


html_last_updated_fmt = '%b %d, %Y'
html_split_index = False
html_sidebars = {
   '**': ['searchbox.html', 'globaltoc.html', 'sourcelink.html'],
}
html_short_title = '%s-%s' % (project, version)


# Mathjax settings
mathjax_path = 'https://cdnjs.cloudflare.com/ajax/libs/mathjax/2.7.5/MathJax.js'
mathjax_config = {
    'extensions': ['tex2jax.js'],
    'jax': ['input/TeX', 'output/SVG'],
    'TeX': {
        'extensions': [
            "AMSmath.js", "AMSsymbols.js", "noErrors.js", "noUndefined.js"],
        'Macros': {
            'tr': ['{\\operatorname{tr}}', 0],
            'diag': ['{\\operatorname{diag}}', 0],
            'abs': ['{\\operatorname{abs}}', 0],
            'pop': ['{\\operatorname{pop}}', 0],
            'ee': ['{\\text{e}}', 0],
            'ii': ['{\\text{i}}', 0],
            'aux': ['{\\text{aux}}', 0],
            'opt': ['{\\text{opt}}', 0],
            'tgt': ['{\\text{tgt}}', 0],
            'init': ['{\\text{init}}', 0],
            'lab': ['{\\text{lab}}', 0],
            'rwa': ['{\\text{rwa}}', 0],
            'bra': ['{\\langle#1\\vert}', 1],
            'ket': ['{\\vert#1\\rangle}', 1],
            'Bra': ['{\\left\\langle#1\\right\\vert}', 1],
            'Braket': ['{\\left\\langle #1\\vphantom{#2} \\mid #2\\vphantom{#1}\\right\\rangle}', 2],
            'Ket': ['{\\left\\vert#1\\right\\rangle}', 1],
            'mat': ['{\\mathbf{#1}}', 1],
            'op': ['{\\hat{#1}}', 1],
            'Op': ['{\\hat{#1}}', 1],
            'dd': ['{\\,\\text{d}}', 0],
            'daggered': ['{^{\\dagger}}', 0],
            'transposed': ['{^{\\text{T}}}', 0],
            'Liouville': ['{\\mathcal{L}}', 0],
            'DynMap': ['{\\mathcal{E}}', 0],
            'identity': ['{\\mathbf{1}}', 0],
            'Norm': ['{\\lVert#1\\rVert}', 1],
            'Abs': ['{\\left\\vert#1\\right\\vert}', 1],
            'avg': ['{\\langle#1\\rangle}', 1],
            'Avg': ['{\\left\langle#1\\right\\rangle}', 1],
            'AbsSq': ['{\\left\\vert#1\\right\\vert^2}', 1],
            'Re': ['{\\operatorname{Re}}', 0],
            'Im': ['{\\operatorname{Im}}', 0],
            'Real': ['{\\mathbb{R}}', 0],
            'Complex': ['{\\mathbb{C}}', 0],
            'Integer': ['{\\mathbb{N}}', 0],
        }
    }
}


# Napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = True
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = True
napoleon_use_param = True
napoleon_use_rtype = True

# -- Extensions to the  Napoleon GoogleDocstring class ---------------------

from sphinx.ext.napoleon.docstring import GoogleDocstring

# first, we define new methods for any new sections and add them to the class
def parse_keys_section(self, section):
    return self._format_fields('Keys', self._consume_fields())
GoogleDocstring._parse_keys_section = parse_keys_section

def parse_attributes_section(self, section):
    return self._format_fields('Attributes', self._consume_fields())
GoogleDocstring._parse_attributes_section = parse_attributes_section

def parse_class_attributes_section(self, section):
    return self._format_fields('Class Attributes', self._consume_fields())
GoogleDocstring._parse_class_attributes_section = parse_class_attributes_section

# we now patch the parse method to guarantee that the the above methods are
# assigned to the _section dict
def patched_parse(self):
    self._sections['keys'] = self._parse_keys_section
    self._sections['class attributes'] = self._parse_class_attributes_section
    self._unpatched_parse()
GoogleDocstring._unpatched_parse = GoogleDocstring._parse
GoogleDocstring._parse = patched_parse


# -- Monkeypatch for instance attribs (sphinx bug #2044) -----------------------


from sphinx.ext.autodoc import (
    ClassLevelDocumenter, InstanceAttributeDocumenter)


def iad_add_directive_header(self, sig):
    ClassLevelDocumenter.add_directive_header(self, sig)


InstanceAttributeDocumenter.add_directive_header = iad_add_directive_header

# -- Options for HTML output ---------------------------------------------------

# on_rtd is whether we are on readthedocs.org, this line of code grabbed from
# docs.readthedocs.org
on_rtd = os.environ.get('READTHEDOCS', None) == 'True'

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
if not on_rtd:  # only import and set the theme if we're building docs locally
    import sphinx_rtd_theme
    html_theme = "sphinx_rtd_theme"
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
#html_theme = 'sphinxdoc'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    'collapse_navigation': True,
    'display_version': True,
    'navigation_depth': 4,
}

# Add any paths that contain custom themes here, relative to this directory.
#html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
#html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
#html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
#html_logo = None

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
#html_favicon = 'favicon.ico'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
#html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
#html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
#html_additional_pages = {}

# If false, no module index is generated.
#html_domain_indices = True

# If false, no index is generated.
#html_use_index = True

# If true, the index is split into individual pages for each letter.
#html_split_index = False

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
#html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
#html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
#html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
#html_file_suffix = None
nbsphinx_prolog = r"""
{% set docname = env.doc2path(env.docname, base='docs') %}

.. only:: html

    .. role:: raw-html(raw)
        :format: html

    :raw-html:`<a href="http://nbviewer.jupyter.org/github/qucontrol/weylchamber/blob/{{ env.config.release }}/{{ docname }}" target="_blank"><img alt="Render on nbviewer" src="https://img.shields.io/badge/render%20on-nbviewer-orange.svg" style="vertical-align:text-bottom"></a>&nbsp;<a href="https://mybinder.org/v2/gh/qucontrol/weylchamber/{{ env.config.release }}?filepath={{ docname }}" target="_blank"><img alt="Launch Binder" src="https://mybinder.org/badge_logo.svg" style="vertical-align:text-bottom"></a>`
"""
# -----------------------------------------------------------------------------
def setup(app):
    app.connect('builder-inited', run_apidoc)



from mpl_toolkits.mplot3d import Axes3D  # required to activate 2D plotting
from mpl_toolkits.mplot3d import proj3d
from matplotlib.patches import FancyArrowPatch

import numpy as np

__all__ = ['WeylChamber']


class WeylChamber():
    """Class for plotting data in the Weyl Chamber

    Class Attributes
    ----------------

    weyl_points: dict
        Dictionary of Weyl chamber point names to (c1, c2, c3) coordinates (in
        units of pi). Each point name is also a class attribute itself
    normal: dict
        Dictionary of Weyl chamber region name to normal vectors for the
        surface that separates the region from the polyhedron of perfect
        entanglers (pointing outwards from the PE's). The three regions are:

        * 'W0': region from the origin point (O) to the PE polyhedron
        * 'W0*': region from the A2 point to the PE polyhedron
        * 'W1': region from the A2 point (SWAP gate) to the PE polyhedron

    anchor: dict
        Dictionary of anchor points for the normal vectors (i.e., an arbitrary
        point on the surface that separates the region specified by the key
        from the perfect entanglers polyhedron

    Attributes
    ----------
    fig_width : float {8.5}
        Figure width, in cm
    fig_height: float {6.0}
        Figure height, in cm
    left_margin: float {0.0}
        Distance from left figure edge to axes, in cm
    bottom_margin: float {0.0}
        Distance from bottom figure edge to axes, in cm
    right_margin: float {0.3}
        Distance from right figure edge to axes, in cm
    top_margin: float {0.0}
        Distance from top figure edge to axes, in cm
    azim: float {-50}
        azimuthal view angle in degrees
    elev: float {-20}
        elevation view angle in degrees
    dpi: int {300}
        Resolution to use when saving to dpi, or showing interactively
    linecolor: str {'black'}
        Color to be used when drawing lines (e.g. the edges of the Weyl
        chamber)
    show_c1_label: bool
        Whether or not to show the c1 axis label
    show_c2_label: bool
        Whether or not to show the c2 axis label
    show_c3_label: bool
        Whether or not to show the c3 axis label
    c1_labelpad: float
        Label padding for c1 label
    c2_labelpad: float
        Label padding for c2 label
    c3_labelpad: float
        Label padding for c3 label
    c1_tickspad: float
        Padding for c1 tick labels
    c2_tickspad: float
        Padding for c2 tick labels
    c3_tickspad: float
        Padding for c2 tick labels
    weyl_edges: list
        List of tuples (point1, point2, foreground) where point1 and point2 are
        keys in `weyl_points`, and foreground is a logical to indicate whether
        the edge is in the background or foreground (depending on the
        perspective of the plot). Describes the lines that make up the Weyl
        chamber.
    weyl_edge_fg_properties: dict
        Properties to be used when drawing foreground weyl_edges
    weyl_edge_bg_properties: dict
        Properties to be used when drawing background weyl_edges
    PE_edges: list
        List of tuples (point1, point2, foreground) where point1 and point2 are
        keys in weyl_points, and foreground is a logical to indicate whether
        the edge is in the background or foreground (depending on the
        perspective of the plot). Desribes the lines that make up the
        polyhedron of perfect entanglers
    PE_edge_fg_properties: dict
        Properties to be used when drawing foreground PE_edges
    PE_edge_bg_properties: dict
        Properties to be used when drawing background PE_edges
    labels: dict
        label names => array (c1, c2, c3) where label should be drawn
    tex_labels: logical {True}
        If True wrap label names in dollar signs to produce a latex string.
    label_properties: dict
        Properties to be used when drawing labels
    z_axis_left: logical {True}
        If True, draw z-axis on the left
    grid: logical {False}
        Show a grid on panes?
    panecolor: None or tuple  {(1.0, 1.0, 1.0, 0.0)}
        Color (r, g, b, alpha) with values in [0,1] for the c1, c2, and c3
        panes
    facecolor: str {'None'}
        Name of color for figure background
    ticklabelsize: float {7}
        font size for tick labels
    full_cube: logical {False}
        if True, draw all three axes in the range [0,1]. This may result in a
        less distorted view of the Weyl chamber
    """
    A1 = np.array((1, 0, 0))
    A2 = np.array((0.5, 0.5, 0))
    A3 = np.array((0.5, 0.5, 0.5))
    O  = np.array((0, 0, 0))
    L  = np.array((0.5, 0, 0))
    M  = np.array((0.75, 0.25, 0))
    N  = np.array((0.75, 0.25, 0.25))
    P  = np.array((0.25, 0.25, 0.25))
    Q  = np.array((0.25, 0.25, 0))
    weyl_points = {'A1' : A1, 'A2' : A2, 'A3' : A3, 'O' : O, 'L' : L,
                   'M' : M, 'N' : N, 'P' : P, 'Q': Q}
    normal = {'W0' : (np.sqrt(2.0)/2.0)*np.array((-1, -1, 0)),
              'W0*': (np.sqrt(2.0)/2.0)*np.array(( 1, -1, 0)),
              'W1' : (np.sqrt(2.0)/2.0)*np.array(( 0,  1, 1))}
    anchor = {'W0': L, 'W0*': L, 'W1': A2}

    def __init__(self):
        self._fig = None
        self._ax = None
        self._artists = None
        self.azim = -50
        self.elev = 20
        self.dpi = 300
        self.fig_width = 8.5
        self.fig_height = 6.0
        self.left_margin = 0.0
        self.bottom_margin =  0.0
        self.right_margin =  0.3
        self.top_margin =  0.0
        self.linecolor = 'black'
        self.show_c1_label = True
        self.show_c2_label = True
        self.show_c3_label = True
        self.c1_labelpad = -9
        self.c2_labelpad = -14
        self.c3_labelpad = -14
        self.c1_tickspad = -6.0
        self.c2_tickspad = -4.0
        self.c3_tickspad = -6.0
        self.weyl_edge_fg_properties = {
                'color':'black', 'linestyle':'-', 'lw':0.5}
        self.weyl_edge_bg_properties = {
                'color':'black', 'linestyle':'--', 'lw':0.5}
        self.PE_edge_fg_properties = {
                'color':'black', 'linestyle':'-', 'lw':0.5}
        self.PE_edge_bg_properties = {
                'color':'black', 'linestyle':'--', 'lw':0.5}
        self.weyl_edges = [
            [ 'O', 'A1', True],
            ['A1', 'A2', True],
            ['A2', 'A3', True],
            ['A3', 'A1', True],
            ['A3',  'O', True],
            [ 'O', 'A2', False]
        ]
        self.PE_edges = [
            ['L',  'N', True],
            ['L',  'P', True],
            ['N',  'P', True],
            ['N', 'A2', True],
            ['N',  'M', True],
            ['M',  'L', False],
            ['Q',  'L', False],
            ['P',  'Q', False],
            ['P', 'A2', False]
        ]
        self.labels = {
            'A_1' : self.A1 + np.array((-0.03, 0.04 , 0.00)),
            'A_2' : self.A2 + np.array((0.01, 0, -0.01)),
            'A_3' : self.A3 + np.array((-0.01, 0, 0)),
            'O'   : self.O  + np.array((-0.025,  0.0, 0.02)),
            'L'   : self.L  + np.array((-0.075, 0, 0.01)),
            'M'   : self.M  + np.array((0.05, -0.01, 0)),
            'N'   : self.N  + np.array((-0.075, 0, 0.009)),
            'P'   : self.P  + np.array((-0.05, 0, 0.008)),
            'Q'   : self.Q  + np.array((0, 0.01, 0.03)),
        }
        self.label_properties = {
            'color': 'black',  'fontsize': 'small'
        }
        self.tex_labels = True
        self.z_axis_left = True
        self.grid = False
        self.panecolor = (1.0, 1.0, 1.0, 0.0)
        self.facecolor = 'None'
        self.ticklabelsize = 7
        self.full_cube = False
        self._scatter = []

    @property
    def figsize(self):
        """Tuple (width, height) of figure size in inches"""
        cm2inch = 0.39370079 # conversion factor cm to inch
        return (self.fig_width*cm2inch, self.fig_height*cm2inch)

    @property
    def fig(self):
        """Return a reference to the figure on which the Weyl chamber has been
        rendered. Undefined unless the `render` method has been called."""
        return self._fig

    @property
    def ax(self):
        """Return a reference to the Axes instance on which the Weyl chamber
        has been rendered. Undefined unless the `render` method has been
        called."""
        return self._ax

    @property
    def artists(self):
        """Return a list of rendered artists. This includes only artists that
        were created as part of a plotting command, not things like the edges
        of the Weyl chamber or the perfect entanglers polyhedron"""
        return self._artists

    def _repr_png_(self):  # pragma: no cover
        from IPython.core.pylabtools import print_figure
        import matplotlib.pyplot as plt
        fig = plt.figure(figsize=self.figsize, dpi=self.dpi)
        ax = fig.add_subplot(111, projection='3d')
        self.render(ax)
        fig_data = print_figure(fig, 'png')
        plt.close(fig)
        return fig_data

    def _repr_svg_(self):  # pragma: no cover
        from IPython.core.pylabtools import print_figure
        import matplotlib.pyplot as plt
        fig = plt.figure(figsize=self.figsize, dpi=self.dpi)
        ax = fig.add_subplot(111, projection='3d')
        self.render(ax)
        fig_data = print_figure(fig, 'svg')
        plt.close(fig)
        return fig_data

    def render(self, ax):
        """Render the Weyl chamber on the given Axes3D object"""
        self._ax = ax
        self._fig = ax.figure
        self._artists = []
        ax.view_init(elev=self.elev, azim=self.azim)
        ax.patch.set_facecolor(self.facecolor)
        if self.panecolor is not None:
            ax.w_xaxis.set_pane_color(self.panecolor)
            ax.w_yaxis.set_pane_color(self.panecolor)
            ax.w_zaxis.set_pane_color(self.panecolor)
        if self.z_axis_left:
            tmp_planes = ax.zaxis._PLANES
            ax.zaxis._PLANES = (tmp_planes[2], tmp_planes[3],
                                tmp_planes[0], tmp_planes[1],
                                tmp_planes[4], tmp_planes[5])
            ax.zaxis.set_rotate_label(False)
            ax.zaxis.label.set_rotation(90)
        ax.grid(self.grid)
        # background lines
        for P1, P2, fg in self.weyl_edges:
            if not fg:
                self._draw_line(ax, P1, P2, zorder=-1,
                               **self.weyl_edge_bg_properties)
        for P1, P2, fg in self.PE_edges:
            if not fg:
                self._draw_line(ax, P1, P2, zorder=-1,
                               **self.PE_edge_bg_properties)
        # scatter plots
        for c1, c2, c3, kwargs in self._scatter:
            self._artists.append(ax.scatter3D(c1, c2, c3, **kwargs))
        pass # plot everything else
        # labels
        for label in self.labels:
            c1, c2, c3 = self.labels[label]
            if self.tex_labels:
                ax.text(c1, c2, c3, "$%s$" % label, **self.label_properties)
            else:
                ax.text(c1, c2, c3, label, **self.label_properties)
        # foreground lines
        for P1, P2, fg in self.weyl_edges:
            if fg:
                self._draw_line(ax, P1, P2, **self.weyl_edge_fg_properties)
        for P1, P2, fg in self.PE_edges:
            if fg:
                self._draw_line(ax, P1, P2, **self.PE_edge_fg_properties)
        ax.set_xlabel(r'$c_1/\pi$')
        ax.set_ylabel(r'$c_2/\pi$')
        ax.set_zlabel(r'$c_3/\pi$')
        if self.full_cube:
            ax.set_xlim(0, 1)
            ax.set_ylim(0, 1)
            ax.set_zlim(0, 1)
        else:
            ax.set_xlim(0, 1)
            ax.set_ylim(0, 0.5)
            ax.set_zlim(0, 0.5)
        if self.show_c1_label:
            ax.set_xlabel(r'$c_1$', labelpad=self.c1_labelpad)
        if self.show_c2_label:
            ax.set_ylabel(r'$c_2$', labelpad=self.c2_labelpad)
        if self.show_c3_label:
            ax.set_zlabel(r'$c_3$', labelpad=self.c3_labelpad)
        ax.set_xlim(0,1)
        ax.set_ylim(0,0.5)
        ax.set_zlim(0,0.5)
        ax.set_xticks([0, 0.25, 0.5, 0.75, 1])
        ax.set_xticklabels(['0', '', '$\pi/2$', '', r'$\pi$'])
        ax.set_yticklabels(['0', '', '', '', '', r'$\pi/2$'])
        ax.set_zticklabels(['0', '', '', '', '', r'$\pi/2$'])
        ax.tick_params(axis='x', pad=self.c1_tickspad)
        ax.tick_params(axis='y', pad=self.c2_tickspad)
        ax.tick_params(axis='z', pad=self.c2_tickspad)
        [t.set_va('center') for t in ax.get_yticklabels()]
        [t.set_ha('left') for t in ax.get_yticklabels()]
        [t.set_va('center') for t in ax.get_zticklabels()]
        [t.set_ha('right') for t in ax.get_zticklabels()]

    def plot(self, fig=None):
        """Generate a plot of the Weyl chamber on the given figure, or create a
        new figure if fig argument is not given.
        """
        if fig is None:
            import matplotlib.pylab as plt
            fig = plt.figure(figsize=self.figsize, dpi=self.dpi)
        w = self.fig_width - (self.left_margin + self.right_margin)
        h = self.fig_height - (self.bottom_margin + self.top_margin)
        pos = [self.left_margin/self.fig_width,
               self.bottom_margin/self.fig_height,
               w/self.fig_width, h/self.fig_height]
        ax = fig.add_axes(pos, projection='3d')
        self.render(ax)

    def scatter(self, c1, c2, c3, **kwargs):
        """Add a scatter plot to the Weyl chamber

        All keyword arguments will be passed to matplotlib scatter3D
        function.
        """
        self._scatter.append((c1, c2, c3, kwargs))

    def add_point(self, c1, c2, c3, scatter_index=0, **kwargs):
        """Add a point to a scatter plot with the given `scatter_index`.

        If there is no existing scatter plot with that index, a new one will be
        created.  The arguments of the scatter plot are updated with the given
        kwargs.
        """
        try:
            c1s, c2s, c3s, kw = self._scatter[scatter_index]
            kw.update(kwargs)
            self._scatter[scatter_index] = (np.append(c1s, [c1, ]),
                                            np.append(c2s, [c2, ]),
                                            np.append(c3s, [c3, ]),
                                            kw)
        except IndexError:
            self._scatter.append((np.array([c1, ]), np.array([c2, ]),
                                  np.array([c3, ]), kwargs))

    def _draw_line(self, ax, origin, end, **kwargs):
        """Draw a line from origin to end onto the given axis. Both origin and
        end must either be 3-tuples, or names of Weyl points. All keyword
        arguments are passed to the `ax.plot` method
        """
        try:
            if origin in self.weyl_points:
                o1, o2, o3 = self.weyl_points[origin]
            else:
                o1, o2, o3 = origin
        except ValueError:  # pragma: nocover
            raise ValueError("origin '%s' is not in weyl_points "
                             "or a list (c1, c2, c3)" % origin)
        try:
            if end in self.weyl_points:
                c1, c2, c3 = self.weyl_points[end]
            else:
                c1, c2, c3 = end
        except ValueError:  # pragma: nocover
            raise ValueError("origin '%s' is not in weyl_points "
                             "or a list (c1, c2, c3)" % origin)
        ax.plot([o1, c1], [o2, c2], [o3, c3], **kwargs)

    def _draw_arrow(self, ax, origin, end, **kwargs):  # pragma: nocover
        # currently undocumented/unused feature, but can be useful when having
        # to mark a point in an error in a publication graphic
        try:
            if origin in self.weyl_points:
                o1, o2, o3 = self.weyl_points[origin]
            else:
                o1, o2, o3 = origin
        except ValueError:  # pragma: nocover
            raise ValueError("origin '%s' is not in weyl_points "
                             "or a list (c1, c2, c3)" % origin)
        try:
            if end in self.weyl_points:
                c1, c2, c3 = self.weyl_points[end]
            else:
                c1, c2, c3 = end
        except ValueError:  # pragma: nocover
            raise ValueError("origin '%s' is not in weyl_points "
                             "or a list (c1, c2, c3)" % origin)
        a = _Arrow3D(
            [o1, c1], [o2, c2], [o3, c3], mutation_scale=10, lw=1.5,
            arrowstyle="-|>", **kwargs)
        ax.add_artist(a)


class _Arrow3D(FancyArrowPatch):  # pragma: no cover
    def __init__(self, xs, ys, zs, *args, **kwargs):
        FancyArrowPatch.__init__(self, (0, 0), (0, 0), *args, **kwargs)
        self._verts3d = xs, ys, zs

    def draw(self, renderer):
        xs3d, ys3d, zs3d = self._verts3d
        xs, ys, zs = proj3d.proj_transform(xs3d, ys3d, zs3d, renderer.M)
        self.set_positions((xs[0], ys[0]), (xs[1], ys[1]))
        FancyArrowPatch.draw(self, renderer)



#: Precision to which to calculate Weyl chamber coordinates or local
#: invariants.
#:
#: Making the precision too high can lead to numerical instabilities, as
#: rounding errors can put a point outside of the Weyl chamber when it should
#: be inside, or into a different region.
DEFAULT_WEYL_PRECISSION = 8



import qutip
import numpy as np
from numpy import pi, sin

from .coordinates import _point_in_PE, weyl_region, c1c2c3, from_magic
from .gates import (
    Qmagic,
    bell_basis as get_bell_basis,
    gate as construct_gate,
    mapped_basis as apply_gate,
)
from ._types import CTuple


__all__ = [
    'project_to_PE',
    'F_PE',
    'concurrence',
    'make_PE_krotov_chi_constructor',
]


#: Dictionary of Weyl chamber region name to normal vectors for the
#: surface that separates the region from the polyhedron of perfect
#: entanglers (pointing outwards from the PE's). The three regions are:
#:
#: * 'W0': region from the origin point (O) to the PE polyhedron
#: * 'W0*': region from the A2 point to the PE polyhedron
#: * 'W1': region from the A2 point (SWAP gate) to the PE polyhedron
_normal = {
    'W0': (np.sqrt(2.0) / 2.0) * np.array((-1, -1, 0)),
    'W0*': (np.sqrt(2.0) / 2.0) * np.array((1, -1, 0)),
    'W1': (np.sqrt(2.0) / 2.0) * np.array((0, 1, 1)),
}


#: Dictionary of anchor points for the normal vectors (i.e., an arbitrary
#: point on the surface that separates the region specified by the key
#: from the perfect entanglers polyhedron
_anchor = {
    'W0': np.array((0.5, 0, 0)),
    'W0*': np.array((0.5, 0, 0)),
    'W1': np.array((0.5, 0.5, 0)),
}


def project_to_PE(c1: float, c2: float, c3: float, check_weyl=True) -> CTuple:
    """Project onto the boundary surface of the perfect entanglers

    Return new tuple (c1', c2', c3') obtained by projecting the given input
    point (c1, c2, c3) onto the closest boundary of the perfect entanglers
    polyhedron. If the input point already is a perfect entangler, it will be
    returned unchanged

    Example:
        >>> from weylchamber.visualize import WeylChamber
        >>> print("%.2f, %.2f, %.2f" % tuple(project_to_PE(*WeylChamber.A3)))
        0.50, 0.25, 0.25
        >>> print("%.3f, %.3f, %.3f" % tuple(project_to_PE(0.5, 0.5, 0.25)))
        0.500, 0.375, 0.125
        >>> print("%.3f, %.3f, %.3f" % tuple(project_to_PE(0.25, 0, 0)))
        0.375, 0.125, 0.000
        >>> print("%.3f, %.3f, %.3f" % tuple(project_to_PE(0.75, 0, 0)))
        0.625, 0.125, 0.000
        >>> print("%.3f, %.3f, %.3f" % tuple(project_to_PE(0.3125, 0.0625, 0.01)))
        0.375, 0.125, 0.010
        >>> print("%.1f, %.1f, %.1f" % tuple(project_to_PE(0.5, 0, 0)))
        0.5, 0.0, 0.0
        >>> print("%.1f, %.1f, %.1f" % tuple(project_to_PE(0.5, 0.2, 0.2)))
        0.5, 0.2, 0.2
        >>> try:
        ...     project_to_PE(1.0, 0.5, 0)
        ... except ValueError as e:
        ...     print(e)
        (1, 0.5, 0) is not in the Weyl chamber
    """
    if _point_in_PE(c1, c2, c3):
        return c1, c2, c3
    else:
        region = weyl_region(c1, c2, c3, check_weyl=check_weyl)
        p = np.array((c1, c2, c3))
        n = _normal[region]
        a = _anchor[region]
        return p - np.inner((p - a), n) * n


def concurrence(c1: float, c2: float, c3: float) -> float:
    """Calculate the concurrence directly from the Weyl Chamber coordinates

    Example:
        >>> import qutip
        >>> from weylchamber.coordinates import c1c2c3
        >>> '%.1f' % concurrence(*c1c2c3(qutip.gates.swap()))
        '0.0'
        >>> '%.1f' % concurrence(*c1c2c3(qutip.gates.cnot()))
        '1.0'
        >>> '%.1f' % concurrence(*c1c2c3(qutip.gates.identity([2, 2])))
        '0.0'
    """
    if ((c1 + c2) >= 0.5) and (c1 - c2 <= 0.5) and ((c2 + c3) <= 0.5):
        # if we're inside the perfect-entangler polyhedron in the Weyl chamber
        # the concurrence is 1 by definition. the "regular" formula gives wrong
        # results in this case.
        return 1
    else:
        c1_c2_c3 = np.array([c1, c2, c3])
        c3_c1_c2 = np.roll(c1_c2_c3, 1)
        m = np.concatenate((c1_c2_c3 - c3_c1_c2, c1_c2_c3 + c3_c1_c2))
        res = np.max(abs(sin(pi * m)))
        if abs(res) < 1e-15:
            return 0
        elif abs(res - 1) < 1e-15:
            return 1
        return res


def F_PE(g1: float, g2: float, g3: float) -> float:
    """Evaluate the Perfect-Entangler Functional

    Example:
        >>> import qutip
        >>> from weylchamber.local_invariants import g1g2g3
        >>> "%.1f" % F_PE(*g1g2g3(qutip.gates.cnot()))
        '0.0'
        >>> "%.1f" % F_PE(*g1g2g3(qutip.gates.identity([2, 2])))
        '2.0'
    """
    return g3 * np.sqrt(g1 ** 2 + g2 ** 2) - g1 + 0.0


def make_PE_krotov_chi_constructor(canonical_basis, unitarity_weight=0):
    r"""Return a constructor for the χ's in a PE optimization.

    Return a `chi_constructor` that determines the boundary condition of the
    backwards propagation in an optimization towards a perfect entangler in
    Krotov's method, based on the foward-propagtion of the Bell states. In
    detail, the function returns a callable function that calculates

    .. math::

        \ket{\chi_{i}}
        =
        \frac{\partial F_{PE}}{\partial \bra{\phi_i}}
        \Bigg|_{\ket{\phi_{i}(T)}}

    for all $i$ with $\ket{\phi_{0}(T)}, ..., \ket{\phi_{3}(T)}$ the forward
    propagated Bell states at final time $T$, cf. Eq. (33b) in Ref. [1].
    $F_{PE}$ is the perfect-entangler functional
    :func:`~weylchamber.perfect_entanglers.F_PE`. For the details of the
    derivative see Appendix G in Ref. [2].

    References:

    [1] `M. H. Goerz, et al., Phys. Rev. A 91, 062307 (2015)
    <https://doi.org/10.1103/PhysRevA.91.062307>`_

    [2] `M. H. Goerz, Optimizing Robust Quantum Gates in Open Quantum Systems.
    PhD thesis, University of Kassel, 2015
    <https://michaelgoerz.net/research/diss_goerz.pdf>`_

    Args:
        canonical_basis (list[qutip.Qobj]): A list of four basis states that
            define the canonical basis $\ket{00}$, $\ket{01}$, $\ket{10}$, and
            $\ket{11}$ of the logical subspace.
        unitarity_weight (float): A weight in [0, 1] that determines how much
            emphasis is placed on  maintaining population in the logical
            subspace.

    Returns:
        callable: a function ``chi_constructor(fw_states_T, **kwargs)`` that
        receives the result of a foward propagation of the Bell states
        (obtained from `canonical_basis` via
        :func:`weylchamber.gates.bell_basis`), and returns a list of statex
        $\ket{\chi_{i}}$ that are the boundary condition for the backward
        propagation in Krotov's method. Positional arguments beyond
        `fw_states_T` are ignored.
    """

    bell_basis = get_bell_basis(canonical_basis)

    w = unitarity_weight
    if w < 0:
        w = 0
    if w > 1:
        w = 1

    def chi_constructor(fw_states_T, **kwargs):
        # *args is ignored, it exists so that the chi_constructor fits the
        # krotov API directly
        UB = construct_gate(bell_basis, fw_states_T)
        A = (Qmagic * _get_a_kl_PE(UB)) / 2

        chis = apply_gate(A, canonical_basis)

        # unitarity corrections
        n = len(fw_states_T)
        chis_out = []
        for i in range(n):
            bell_proj = qutip.Qobj(np.zeros(shape=fw_states_T[i].shape))
            for j in range(n):
                bell_proj += bell_basis[j].overlap(fw_states_T[i])            \
                             * bell_basis[j]
            chis_out.append((1.0-w) * chis[i] + 0.25*w * bell_proj)
        return chis_out

    return chi_constructor


def _get_a_kl_PE(UB):
    """Return the 4×4 `A_kl` coefficient matrix (:class:`qutip.Qobj`)
    for the perfect-entanglers functional, for a given gate `UB` in the Bell
    basis.
    """

    # Compute auxiliary scalar quantities
    detU = _cmat4_det(UB)
    rcdetU1 = 1.0 / detU
    mU = UB.trans() * UB
    tr_mU = mU.tr()
    tr_mU_sq = (mU*mU).tr()
    g3 = 0.25 * (tr_mU**2 - tr_mU_sq)
    Re2Tr = np.real(tr_mU)**2 / 16.0 # This contains already the 1/16 factor
    Im2Tr = np.imag(tr_mU)**2 / 16.0 # This contains already the 1/16 factor
    Tr2 = tr_mU**2 / 16.0            # This contains already the 1/16 factor

    # `alpha` and `beta` are the real and imaginary part of `UB`, respectively
    alpha = np.zeros(shape=(4,4))
    beta  = np.zeros(shape=(4,4))
    for i in range(4):
        for j in range(4):
            alpha[i,j] = np.real(UB[i,j])
            beta[i,j]  = np.imag(UB[i,j])
    alpha = qutip.Qobj(alpha)
    beta  = qutip.Qobj(beta)

    # The c1c2c3 function expects `UB` to be given in the canonical basis, thus
    # we have to convert it from the Bell basis on input
    c1, c2, c3 = c1c2c3(from_magic(UB))

    drcdetU1dAlpha = np.zeros(shape=(4,4), dtype=complex)
    drcdetU1dBeta  = np.zeros(shape=(4,4), dtype=complex)
    dg3dAlpha      = np.zeros(shape=(4,4), dtype=complex)
    dg3dBeta       = np.zeros(shape=(4,4), dtype=complex)
    dRe2TrdAlpha   = np.zeros(shape=(4,4))
    dRe2TrdBeta    = np.zeros(shape=(4,4))
    dIm2TrdAlpha   = np.zeros(shape=(4,4))
    dIm2TrdBeta    = np.zeros(shape=(4,4))
    dTr2dAlpha     = np.zeros(shape=(4,4), dtype=complex)
    dTr2dBeta      = np.zeros(shape=(4,4), dtype=complex)

    # Compute auxiliary matrix quantities
    for a in range(4):
        for b in range(4):

            drcdetU1dAlpha[a,b] = - _dDetUdAlpha(UB,a,b) / detU**2
            drcdetU1dBeta[a,b]  = - _dDetUdBeta(UB,a,b)  / detU**2

            for k in range(4):
                for i in range(4):
                    dg3dAlpha[a,b] = dg3dAlpha[a,b]                           \
                                 +       alpha[a,b] * alpha[k,i] * alpha[k,i] \
                                 -       alpha[a,b] * beta[k,i]  * beta[k,i]  \
                                 - 2.0 * beta[a,b]  * alpha[k,i] * beta[k,i]  \
                                 -       alpha[k,i] * alpha[a,i] * alpha[k,b] \
                                 +       alpha[k,b] * beta[a,i]  * beta[k,i]  \
                                 + 2.0 * beta[k,b]  * alpha[a,i] * beta[k,i]
                    dg3dBeta[a,b]  = dg3dBeta[a,b]                            \
                                 +       beta[a,b]  * beta[k,i]  * beta[k,i]  \
                                 -       beta[a,b]  * alpha[k,i] * alpha[k,i] \
                                 - 2.0 * alpha[a,b] * alpha[k,i] * beta[k,i]  \
                                 -       beta[k,i]  * beta[a,i]  * beta[k,b]  \
                                 +       beta[k,b]  * alpha[a,i] * alpha[k,i] \
                                 + 2.0 * alpha[k,b] * beta[a,i]  * alpha[k,i]
                    dg3dAlpha[a,b] = dg3dAlpha[a,b] + 1j * (                  \
                                 +       beta[a,b]  * alpha[k,i] * alpha[k,i] \
                                 -       beta[a,b]  * beta[k,i]  * beta[k,i]  \
                                 + 2.0 * alpha[a,b] * alpha[k,i] * beta[k,i]  \
                                 -       alpha[a,i] * alpha[k,i] * beta[k,b]  \
                                 -       alpha[k,b] * alpha[k,i] * beta[a,i]  \
                                 -       alpha[k,b] * alpha[a,i] * beta[k,i]  \
                                 +       beta[a,i]  * beta[k,i]  * beta[k,b]  \
                                 )
                    dg3dBeta[a,b]  = dg3dBeta[a,b] + 1j * (                   \
                                 -       alpha[a,b] * beta[k,i]  * beta[k,i]  \
                                 +       alpha[a,b] * alpha[k,i] * alpha[k,i] \
                                 - 2.0 * beta[a,b]  * beta[k,i]  * alpha[k,i] \
                                 +       beta[a,i]  * beta[k,i]  * alpha[k,b] \
                                 +       beta[k,b]  * beta[k,i]  * alpha[a,i] \
                                 +       beta[k,b]  * beta[a,i]  * alpha[k,i] \
                                 -       alpha[a,i] * alpha[k,i] * alpha[k,b] \
                                 )

            for k in range(4):
                for i in range(4):
                    dTr2dAlpha[a,b] = dTr2dAlpha[a,b]                         \
                                + 0.25 * alpha[a,b] * alpha[k,i] * alpha[k,i] \
                                - 0.25 * alpha[a,b] * beta[k,i]  * beta[k,i]  \
                                - 0.50 * beta[a,b]  * alpha[k,i] * beta[k,i]
                    dTr2dBeta[a,b]  = dTr2dBeta[a,b]                          \
                                + 0.25 * beta[a,b]  * beta[k,i]  * beta[k,i]  \
                                - 0.25 * beta[a,b]  * alpha[k,i] * alpha[k,i] \
                                - 0.50 * alpha[a,b] * beta[k,i]  * alpha[k,i]
                    dTr2dAlpha[a,b] = dTr2dAlpha[a,b] + 1j * (                \
                                + 0.25 * beta[a,b]  * alpha[k,i] * alpha[k,i] \
                                - 0.25 * beta[a,b]  * beta[k,i]  * beta[k,i]  \
                                + 0.50 * alpha[a,b] * alpha[k,i] * beta[k,i]  \
                                )
                    dTr2dBeta[a,b]  = dTr2dBeta[a,b] + 1j * (                 \
                                - 0.25 * alpha[a,b] * beta[k,i]  * beta[k,i]  \
                                + 0.25 * alpha[a,b] * alpha[k,i] * alpha[k,i] \
                                - 0.50 * beta[a,b]  * beta[k,i]  * alpha[k,i] \
                                )

            for k in range(4):
                for i in range(4):
                    dRe2TrdAlpha[a,b] = dRe2TrdAlpha[a,b] + np.real(          \
                                  0.25 * alpha[a,b] * alpha[k,i] * alpha[k,i] \
                                - 0.25 * alpha[a,b] * beta[k,i] * beta[k,i]   \
                                )
                    dRe2TrdBeta[a,b]  = dRe2TrdBeta[a,b] + np.real(           \
                                   0.25 * beta[a,b] * beta[k,i] * beta[k,i]   \
                                 - 0.25 * beta[a,b] * alpha[k,i] * alpha[k,i] \
                                 )
                    dIm2TrdAlpha[a,b] = dIm2TrdAlpha[a,b] + np.real(          \
                                    0.50 * alpha[k,i] * beta[a,b] * beta[k,i] \
                                    )
                    dIm2TrdBeta[a,b]  = dIm2TrdBeta[a,b] + np.real(           \
                                   0.50 * beta[k,i] * alpha[a,b] * alpha[k,i] \
                                   )

    # Construct the a_kl coefficients with the previously computed quantities
    a_kl_coeffs = np.zeros(shape=(4,4), dtype=complex)
    for l in range(4):
        for k in range(4):
            a_kl_coeffs[k,l] =                                                \
            np.real(                                                          \
              - drcdetU1dAlpha[k,l] * g3 * Re2Tr                              \
              - rcdetU1 * dg3dAlpha[k,l] * Re2Tr                              \
              - rcdetU1 * g3 * dRe2TrdAlpha[k,l]                              \
              - drcdetU1dAlpha[k,l] * g3 * Im2Tr                              \
              - rcdetU1 * dg3dAlpha[k,l] * Im2Tr                              \
              - rcdetU1 * g3 * dIm2TrdAlpha[k,l]                              \
              + drcdetU1dAlpha[k,l] * Tr2                                     \
              + rcdetU1 * dTr2dAlpha[k,l])                                    \
            + 1j *                                                            \
            np.real(                                                          \
              - drcdetU1dBeta[k,l] * g3 * Re2Tr                               \
              - rcdetU1 * dg3dBeta[k,l] * Re2Tr                               \
              - rcdetU1 * g3 * dRe2TrdBeta[k,l]                               \
              - drcdetU1dBeta[k,l] * g3 * Im2Tr                               \
              - rcdetU1 * dg3dBeta[k,l] * Im2Tr                               \
              - rcdetU1 * g3 * dIm2TrdBeta[k,l]                               \
              + drcdetU1dBeta[k,l] * Tr2                                      \
              + rcdetU1 * dTr2dBeta[k,l])

    if (c2 + c3 > 0.5):
        # UB is in the W1 region of the Weyl chamber (between the PE polyhedron
        # and the SWAP gate at the A3 point). In this region, the PE-functional
        # has the wrong sign (cf. Fig 6.1 in Goerz, PhD Thesis, Kassel).
        a_kl_coeffs = -a_kl_coeffs
        # Without this corrections, gates would be pushed towards [SWAP]

    return a_kl_coeffs


def _cmat4_det(m):
    """Calculates the complex determinant of a 4x4 matrix.

    Args:
        m (qutip.Qobj): A 4x4 quantum gate for a two-qubit system

    Returns:
        float: The determinant $\det(m)$
    """
    d =   m[0,3]*m[1,2]*m[2,1]*m[3,0] - m[0,2]*m[1,3]*m[2,1]*m[3,0]           \
        - m[0,3]*m[1,1]*m[2,2]*m[3,0] + m[0,1]*m[1,3]*m[2,2]*m[3,0]           \
        + m[0,2]*m[1,1]*m[2,3]*m[3,0] - m[0,1]*m[1,2]*m[2,3]*m[3,0]           \
        - m[0,3]*m[1,2]*m[2,0]*m[3,1] + m[0,2]*m[1,3]*m[2,0]*m[3,1]           \
        + m[0,3]*m[1,0]*m[2,2]*m[3,1] - m[0,0]*m[1,3]*m[2,2]*m[3,1]           \
        - m[0,2]*m[1,0]*m[2,3]*m[3,1] + m[0,0]*m[1,2]*m[2,3]*m[3,1]           \
        + m[0,3]*m[1,1]*m[2,0]*m[3,2] - m[0,1]*m[1,3]*m[2,0]*m[3,2]           \
        - m[0,3]*m[1,0]*m[2,1]*m[3,2] + m[0,0]*m[1,3]*m[2,1]*m[3,2]           \
        + m[0,1]*m[1,0]*m[2,3]*m[3,2] - m[0,0]*m[1,1]*m[2,3]*m[3,2]           \
        - m[0,2]*m[1,1]*m[2,0]*m[3,3] + m[0,1]*m[1,2]*m[2,0]*m[3,3]           \
        + m[0,2]*m[1,0]*m[2,1]*m[3,3] - m[0,0]*m[1,2]*m[2,1]*m[3,3]           \
        - m[0,1]*m[1,0]*m[2,2]*m[3,3] + m[0,0]*m[1,1]*m[2,2]*m[3,3]
    return d


def _dDetUdAlpha(U,a,b):
    """Calculates

    .. math::

        \partial \det(U) / \partial \\alpha_{a,b} = \det(U'_{a,b})

    where $U'$ is constructed from $U$ by replacing the row $a$ with the unit
    vector $b$ and $\\alpha$ is the real part of $U$.

    Args:
        U (qutip.Qobj): A 4x4 quantum gate for a two-qubit system
        a (integer): An integer specifying the derivative
        b (integer): An integer specifying the derivative

    Returns:
        float: The derivative as specified by $a$ and $b$
    """
    U_prime = np.zeros(shape=(4,4), dtype=complex)
    for i in range(4):
        for j in range(4):
            if i==a:
                U_prime[i,j] = 0
            else:
                U_prime[i,j] = U[i,j]
    U_prime[a,b] = 1
    return _cmat4_det(qutip.Qobj(U_prime))


def _dDetUdBeta(U,a,b):
    """Calculates

    .. math::

        \partial \det(U) / \partial \\beta_{a,b} = \det(U'_{a,b})

    where $U'$ is constructed from $U$ by replacing the row $a$ with the unit
    vector $b$ and $\\beta$ is the imaginary part of $U$.

    Args:
        U (qutip.Qobj): A 4x4 quantum gate for a two-qubit system
        a (integer): An integer specifying the derivative
        b (integer): An integer specifying the derivative

    Returns:
        float: The derivative as specified by $a$ and $b$
    """
    return 1j*_dDetUdAlpha(U,a,b)



import numpy as np
import qutip
from qutip import Qobj

from .coordinates import to_magic, from_magic, c1c2c3, canonical_gate
from ._types import Gate, Tuple

__all__ = ['cartan_decomposition']


def cartan_decomposition(U: Gate) -> Tuple[Qobj, Qobj, Qobj]:
    r"""Calculate the Cartan Decomposition of the given U in U(4)

    .. math::

        \Op{U} = \Op{k}_1  \Op{A}  \Op{k}_2

    up to a global phase factor $(\det \Op{U})^{\frac{1}{4}}$.

    Args:
        U: Two-qubit quantum gate. Must be unitary.

    Returns:
        tuple $(\Op{k}_1, \Op{A}, \Op{k}_2)$ where
        $\Op{k}_1$ is the left local operations in SU(2) x SU(2),
        $\Op{A}$ is non-local operations, in SU(4), and $\Op{k}_2$
        is the right local operations in SU(2) x SU(2).

    Notes:

        If you are working with a logical subspace, you should unitarize U
        before calculating the Cartan decomposition

    References:

        * D. Reich. Optimising the nonlocal content of a two-qubit gate.
          Diploma Thesis. FU Berlin, 2010. Appendix E

        * Zhang et al. PRA 67, 042313 (2003)
    """
    if isinstance(U, qutip.Qobj):
        U = U.full()
    U = np.array(U)                      # in U(4)
    Utilde = U / np.linalg.det(U)**0.25  # U in SU(4)

    found_branch = False
    # The fourth root has four branches; the correct solution could be in
    # any one of them
    for branch in range(4):

        UB = to_magic(Utilde).full()  # in Bell basis
        m = UB.T @ UB

        # The F-matrix can be calculated according to Eq (21) in PRA 67, 042313
        # It is a diagonal matrix containing the squares of the eigenvalues of
        # m
        c1, c2, c3 = c1c2c3(Utilde)
        F1 = np.exp(np.pi * 0.5j * (+c1 - c2 + c3))
        F2 = np.exp(np.pi * 0.5j * (+c1 + c2 - c3))
        F3 = np.exp(np.pi * 0.5j * (-c1 - c2 - c3))
        F4 = np.exp(np.pi * 0.5j * (-c1 + c2 + c3))
        Fd = np.array([F1, F2, F3, F4])
        F = np.array(np.diag(Fd))

        # Furthermore, we have Eq (22), giving the eigen-decomposition of the
        # matrix m. This gives us the matrix O_2.T of the eigenvectors of m
        Fsq, O_2_transposed = np.linalg.eig(m)

        ord1 = np.argsort(np.angle(Fd**2))  # sort according to complex phase
        ord2 = np.argsort(np.angle(Fsq))    # ... (absolute value is 1)
        diff = np.sum(np.abs((Fd**2)[ord1] - Fsq[ord2]))
        # Do Fd**2 and Fsq contain the same values (irrespective of order)?
        if diff < 1.0e-12:
            found_branch = True
            break
        else:
            Utilde *= 1.0j

    # double check that we managed to find a branch (just to be 100% safe)
    assert(found_branch), \
        "Couldn't find correct branch of fourth root in mapping U(4) -> SU(4)"

    # Getting the entries of F from Eq (21) instead of by taking the square
    # root of Fsq has the benefit that we don't have to worry about whether we
    # need to take the positive or negative root.
    # However, we do need to make sure that the eigenstates are ordered to
    # correspond to F1, F2, F3, F4
    # After reordering, we need to transpose to get O_2 itself
    reordered = np.array(np.zeros((4, 4)), dtype=np.complex128)
    order = []
    for i in range(4):
        for j in range(4):
            if (abs(Fd[i]**2 - Fsq[j]) < 1.0e-12):
                if j not in order:
                    order.append(j)
    assert len(order) == 4, "Couldn't order O_2"  # should not happen
    # Reorder using the order we just figured out, and transpose
    for i in range(4):
        reordered[:, i] = O_2_transposed[:, order[i]]
    O_2 = reordered.transpose()

    # Now that we have O_2 and F, completing the Cartan decomposition is
    # straightforward, following along Appendix E of Daniel's thesis
    k2 = from_magic(O_2)
    O_1 = UB @ O_2.transpose() @ F.conjugate().transpose()
    k1 = from_magic(O_1)
    A = canonical_gate(c1, c2, c3)

    # Check our results
    assert(np.max(np.abs(O_1 @ O_1.transpose() - np.identity(4))) < 1.0e-12), \
        "O_1 not orthogonal"
    assert(np.max(np.abs(O_2 @ O_2.transpose() - np.identity(4))) < 1.0e-12), \
        "O_2 not orthogonal"
    err = (k1 * A * k2 - qutip.Qobj(Utilde, dims=[[2, 2], [2, 2]])).norm()
    assert(err < 1e-12), "Cartan Decomposition Failed"
    return k1, A, k2



"""Calculation of Weyl chamber coordinates"""

import numpy as np
from numpy import less, greater
from numpy import logical_and as And
from numpy import logical_or as Or
from numpy import less_equal as less_eq
from numpy import greater_equal as greater_eq
import qutip

from .gates import Qmagic, SxSx, SySy, SzSz
from .prec import DEFAULT_WEYL_PRECISSION
from ._types import Gate, CTuple

__all__ = [
    'c1c2c3', 'canonical_gate', 'from_magic', 'point_in_region',
    'point_in_weyl_chamber', 'random_gate', 'random_weyl_point', 'to_magic',
    'weyl_region']


def c1c2c3(U: Gate, ndigits=DEFAULT_WEYL_PRECISSION) -> CTuple:
    """Calculate Weyl chamber coordinates $(c_1, c_2, c_3)$

    Given U (in canonical basis), calculate the Weyl Chamber coordinates
    $(c_1, c_2, c_3)$, in units of π.

    In order to facility numerical stability, the resulting coordinates are
    rounded to the given precision (cf. `ndigits` parameter of the built-in
    `round` function). Otherwise, rounding errors would likely to result in
    points that are not in the Weyl chamber, e.g. (0.1, 0.0, 1.0e-13)

    Algorithm from Childs et al., PRA 68, 052311 (2003).

    Example:
        >>> print("%.2f %.2f %.2f" % c1c2c3(qutip.gates.cnot()))
        0.50 0.00 0.00
    """
    U = qutip.Qobj(U, dims=[[2, 2], [2, 2]])
    if U.shape != (4, 4):
        raise ValueError("Gates must have a 4×4 shape")
    U_tilde = SySy * U.trans() * SySy
    ev = np.linalg.eigvals(
        ((U * U_tilde).full())/np.sqrt(complex(np.linalg.det(U.full()))))
    two_S = np.angle(ev) / np.pi
    for i in range(len(two_S)):
        if two_S[i] <= -0.5:
            two_S[i] += 2.0
    S = np.sort(two_S / 2.0)[::-1]  # sort decreasing
    n = int(round(sum(S)))
    S -= np.r_[np.ones(n), np.zeros(4-n)]
    S = np.roll(S, -n)
    M = np.array([[1, 1, 0], [1, 0, 1], [0, 1, 1]])
    c1, c2, c3 = np.dot(M, S[:3])
    if c3 < 0:
        c1 = 1 - c1
        c3 = -c3
    return (round(c1+0.0, ndigits),  # adding 0.0 turns -0.0 into +0.0
            round(c2+0.0, ndigits),
            round(c3+0.0, ndigits))


def point_in_weyl_chamber(
        c1: float, c2: float, c3: float, raise_exception=False) -> bool:
    """Check if the coordinates $(c_1, c_2, c_3)$ are inside the Weyl chamber

    Examples:

        >>> BGATE = qutip.gates.berkeley()
        >>> point_in_weyl_chamber(*c1c2c3(BGATE))
        True
        >>> point_in_weyl_chamber(*c1c2c3(qutip.gates.identity([2, 2])))
        True

        The coordinates may also be array-like, in which case a boolean numpy
        array is returned.

        >>> res = point_in_weyl_chamber(
        ...     [0.0,0.5,0.8], [1.0,0.25,0.0], [1.0,0.25,0.0])
        >>> assert np.all(res == np.array([False,  True,  True]))

        If `raise_exception` is True, raise an :exc:`ValueError` if any values
        are outside the Weyl chamber.

        >>> try:
        ...     point_in_weyl_chamber(1.0, 0.5, 0, raise_exception=True)
        ... except ValueError as e:
        ...     print(e)
        (1, 0.5, 0) is not in the Weyl chamber
    """
    result = Or(
        And(And(
          less(c1, 0.5), less_eq(c2, c1)), less_eq(c3, c2)),
        And(And(
          greater_eq(c1, 0.5), less_eq(c2, 1.0-np.array(c1))), less_eq(c3, c2))
    )
    if raise_exception:
        if not np.all(result):
            if np.isscalar(c1):  # assume c2, c3 are scalar, too
                raise ValueError(
                    "(%g, %g, %g) is not in the Weyl chamber" % (c1, c2, c3))
            else:
                raise ValueError(
                    "Not all values (c1, c2, c3) are in the Weyl chamber")
    return result


def point_in_region(
        region: str, c1: float, c2: float, c3: float,
        check_weyl=False) -> bool:
    """Check if $(c_1, c_2, c_3)$ are in the given region of the Weyl chamber

    The regions are 'W0' (between origin O and perfect entanglers polyhedron),
    'W0*' (between point A1 and perfect entangler polyhedron), 'W1' (between A3
    point and perfect entanglers polyhedron), and 'PE' (inside perfect
    entanglers polyhedron)

    If the `check_weyl` parameter is given a True, raise a ValueError for any
    points outside of the Weyl chamber

    Examples:
        >>> from weylchamber.visualize import WeylChamber
        >>> point_in_region('W0', *WeylChamber.O)
        True
        >>> point_in_region('W0', 0.2, 0.05, 0.0)
        True
        >>> point_in_region('W0', *WeylChamber.L)
        False
        >>> point_in_region('W0', *WeylChamber.Q)
        False
        >>> point_in_region('PE', *WeylChamber.Q)
        True
        >>> point_in_region('W0*', *WeylChamber.A1)
        True
        >>> point_in_region('W0*', 0.8, 0.1, 0.1)
        True
        >>> point_in_region('W1', *WeylChamber.A3)
        True
        >>> point_in_region('W1', 0.5, 0.4, 0.25)
        True
        >>> point_in_region('W1', 0.5, 0.25, 0)
        False
        >>> point_in_region('PE', 0.5, 0.25, 0)
        True

        The function may be also applied against arrays::

        >>> res = point_in_region('W1', [0.5,0.5], [0.4,0.25], [0.25,0.0])
        >>> assert np.all(res == np.array([ True, False]))
    """
    regions = ['W0', 'W0*', 'W1', 'PE']
    if region == 'PE':
        return _point_in_PE(c1, c2, c3, check_weyl=check_weyl)
    else:
        in_weyl = point_in_weyl_chamber(c1, c2, c3, raise_exception=check_weyl)
        c1 = np.array(c1)
        c2 = np.array(c2)
        c3 = np.array(c3)
        if region == 'W0':
            return And(in_weyl, less(c1+c2, 0.5))
        elif region == 'W0*':
            return And(in_weyl, greater(c1-c2, 0.5))
        elif region == 'W1':
            return And(in_weyl, greater(c2+c3, 0.5))
        else:
            raise ValueError("region %s is not in %s" % (region, regions))


def _point_in_PE(c1, c2, c3, check_weyl=False):
    """Return True if the coordinates c1, c2, c3 are inside the
    perfect-entangler polyhedron

    >>> BGATE = qutip.gates.berkeley()
    >>> _point_in_PE(*c1c2c3(BGATE))
    True
    >>> _point_in_PE(*c1c2c3(qutip.gates.identity(4)))
    False

    >>> res = _point_in_PE([0.0, 0.5, 0.8], [1.0, 0.25, 0.0], [1.0, 0.25, 0.0])
    >>> assert np.all(res == np.array([False,  True, False]))
    """
    in_weyl = point_in_weyl_chamber(c1, c2, c3, raise_exception=check_weyl)
    c1 = np.array(c1); c2 = np.array(c2); c3 = np.array(c3)
    return And(in_weyl,
        And(And(
            greater_eq(c1+c2, 0.5), less_eq(c1-c2, 0.5)), less_eq(c2+c3, 0.5)
        )
    )


def weyl_region(c1: float, c2: float, c3: float, check_weyl=True) -> str:
    """Return the region of the Weyl chamber the given point is in.

    Returns:
        One of 'W0', 'W0*', 'W1', 'PE'

    Examples:

        >>> from weylchamber.visualize import WeylChamber
        >>> print(weyl_region(*WeylChamber.O))
        W0
        >>> print(weyl_region(*WeylChamber.A1))
        W0*
        >>> print(weyl_region(*WeylChamber.A3))
        W1
        >>> print(weyl_region(*WeylChamber.L))
        PE
        >>> print(weyl_region(0.2, 0.05, 0.0))
        W0
        >>> print(weyl_region(0.8, 0.1, 0.1))
        W0*
        >>> print(weyl_region(0.5, 0.25, 0))
        PE
        >>> print(weyl_region(0.5, 0.4, 0.25))
        W1
        >>> try:
        ...     weyl_region(1.0, 0.5, 0)
        ... except ValueError as e:
        ...     print(e)
        (1, 0.5, 0) is not in the Weyl chamber
        >>> print(weyl_region(1.0, 0.1, 0, check_weyl=False))
        W0*

    Only scalar values are accepted for c1, c2, c3
    """
    point_in_weyl_chamber(c1, c2, c3, raise_exception=check_weyl)
    if c1+c2 < 0.5:
        return 'W0'
    elif c1-c2 > 0.5:
        return 'W0*'
    elif c2+c3 > 0.5:
        return 'W1'
    else:
        return 'PE'


def to_magic(A: Gate) -> qutip.Qobj:
    """Convert `A` from the canonical basis to the the "magic" Bell basis"""
    if A.shape != (4, 4):
        raise ValueError("Gates must have a 4×4 shape")
    return Qmagic.dag() * qutip.Qobj(A, dims=[[2, 2], [2, 2]]) * Qmagic


def from_magic(A: Gate) -> qutip.Qobj:
    """ The inverse of :func:`.to_magic`"""
    if A.shape != (4, 4):
        raise ValueError("Gates must have a 4×4 shape")
    return Qmagic * qutip.Qobj(A, dims=[[2, 2], [2, 2]]) * Qmagic.dag()


def random_weyl_point(region=None) -> CTuple:
    """Return a random point $(c_1, c_2, c_3)$ in the Weyl chamber (units of π)

    If region is given in ['W0', 'W0*', 'W1', 'PE'], the point will be in the
    specified region of the Weyl chamber

    Example:
        >>> c1, c2, c3 = random_weyl_point()
        >>> point_in_weyl_chamber(c1, c2, c3)
        True
        >>> c1, c2, c3 = random_weyl_point(region='PE')
        >>> point_in_region('PE', c1, c2, c3)
        True
        >>> c1, c2, c3 = random_weyl_point(region='W0')
        >>> point_in_region('W0', c1, c2, c3)
        True
        >>> c1, c2, c3 = random_weyl_point(region='W0*')
        >>> point_in_region('W0*', c1, c2, c3)
        True
        >>> c1, c2, c3 = random_weyl_point(region='W1')
        >>> point_in_region('W1', c1, c2, c3)
        True
    """
    while True:
        c1 = np.random.rand()
        c2 = 0.5*np.random.rand()
        c3 = 0.5*np.random.rand()
        if point_in_weyl_chamber(c1, c2, c3):
            if region is None:
                return c1, c2, c3
            else:
                if point_in_region(region, c1, c2, c3):
                    return c1, c2, c3


def _U2(phi, theta, phi1, phi2):
    r"""Return a unitary gate as numpy array using the parametrization

    .. math::

        U = e^{i \phi} \begin{bmatrix}
             \cos\theta e^{ i\phi_1}  & \sin\theta e^{ i\phi_2}\\
            -\sin\theta e^{-i\phi_2}  & \cos\theta e^{-i\phi_1}\\
            \end{bmatrix}
    """
    from numpy import exp, sin, cos
    return exp(1j*phi) * np.array([
        [  cos(theta) * exp( 1j*phi1), sin(theta) * exp( 1j*phi2) ],
        [ -sin(theta) * exp(-1j*phi2), cos(theta) * exp(-1j*phi1) ]])


def _SQ_unitary(
        phi_left, theta_left, phi1_left, phi2_left, phi_right, theta_right,
        phi1_right, phi2_right):
    """Return a non-entangling two-qubit gate (a two-qubit gate locally
    equivalent to the identity)"""
    Id = np.identity(2)
    return qutip.Qobj(
        np.kron(_U2(phi_left, theta_left, phi1_left, phi2_left), Id).dot(
            np.kron(Id, _U2(phi_right, theta_right, phi1_right, phi2_right))),
        dims=[[2, 2], [2, 2]])


def canonical_gate(c1: float, c2: float, c3: float) -> qutip.Qobj:
    r"""Return the canonical gate for the given $(c_1, c_2, c_3)$

    The canonical gate is defined as

    .. math::

        \Op{A} = e^{i \frac{\pi}{2} \left(
            c_1 \Op{\sigma}_x \Op{\sigma}_x +
            c_2 \Op{\sigma}_y \Op{\sigma}_y +
            c_3 \Op{\sigma}_z \Op{\sigma}_z \right)}

    Example:
        >>> gate = qutip.Qobj(
        ...     [[0.707107+0.000000j, 0.000000+0.000000j, 0.000000+0.000000j, 0.000000+0.707107j],
        ...      [0.000000+0.000000j, 0.707107+0.000000j, 0.000000+0.707107j, 0.000000+0.000000j],
        ...      [0.000000+0.000000j, 0.000000+0.707107j, 0.707107+0.000000j, 0.000000+0.000000j],
        ...      [0.000000+0.707107j, 0.000000+0.000000j, 0.000000+0.000000j, 0.707107+0.000000j]],
        ... dims=[[2,2], [2,2]])
        >>> U = canonical_gate(0.5,0,0)
        >>> assert (U - gate).norm() < 1e-6
        >>> assert np.max(np.abs(
        ...     np.array(c1c2c3(U)) - np.array([0.5, 0, 0]))) < 1e-15
        """
    return (np.pi*0.5j * (c1 * SxSx + c2 * SySy + c3 * SzSz)).expm()


def random_gate(region=None) -> qutip.Qobj:
    """Return a random two-qubit gate

    If region is not None, the gate will be in the specified region of the Weyl
    chamber.  The following region names are allowed:

    * 'SQ': single qubit gates, consisting of the Weyl chamber points O, A2
    * 'PE': polyhedron of perfect entanglers
    * 'W0': region between point O and the PE polyhedron
    * 'W0*': region between point A2 and the PE polyhedron
    * 'W1': region between point A3 ([SWAP]) and the PE polyhedron
    """
    if region == 'SQ':
        p = 2.0 * np.pi * np.random.random(8)
        return _SQ_unitary(*p)
    else:
        A = canonical_gate(*random_weyl_point(region=region))  # Qobj
        p = 2.0 * np.pi * np.random.random(16)
        return _SQ_unitary(*p[:8]) * A * _SQ_unitary(*p[8:])



import warnings

from weylchamber import random_gate, concurrence, c1c2c3, cartan_decomposition
import qutip
import numpy as np

import pytest


@pytest.mark.xfail
def test_cartan_decomposition_random():
    U = random_gate()
    phasefactor = np.linalg.det(U.full())**0.25
    assert abs(np.linalg.det(U.full()) - 1) < 1e-15
    C = concurrence(*c1c2c3(U))
    k1, A, k2 = cartan_decomposition(U)
    assert ((phasefactor * k1 * A * k2) - U).norm() < 1e-15
    assert concurrence(*c1c2c3(k1)) == 0
    assert concurrence(*c1c2c3(k2)) == 0
    assert abs(concurrence(*c1c2c3(A)) - C) < 1e-15


def test_cartan_decomposition_cnot():
    warnings.filterwarnings(
        'ignore', message='the matrix subclass is not the recommended way')
    U = qutip.gates.cnot()
    phasefactor = np.linalg.det(U.full())**0.25
    C = concurrence(*c1c2c3(U))
    assert C == 1
    k1, A, k2 = cartan_decomposition(U)
    assert ((phasefactor * k1 * A * k2) - U).norm() < 1e-15
    assert concurrence(*c1c2c3(k1)) == 0
    assert concurrence(*c1c2c3(k2)) == 0
    assert abs(concurrence(*c1c2c3(A)) - C) < 1e-15



import qutip
from weylchamber import concurrence, random_gate, weyl_region, c1c2c3

def test_random_gate():
    G = random_gate("SQ")
    assert isinstance(G, qutip.Qobj)
    assert G.dims == [[2, 2], [2, 2]]
    assert concurrence(*c1c2c3(G)) == 0
    G = random_gate("PE")
    assert isinstance(G, qutip.Qobj)
    assert G.dims == [[2, 2], [2, 2]]
    assert concurrence(*c1c2c3(G)) == 1
    assert weyl_region(*c1c2c3(random_gate("W0"))) == 'W0'
    assert weyl_region(*c1c2c3(random_gate("W0*"))) == 'W0*'
    assert weyl_region(*c1c2c3(random_gate("W1"))) == 'W1'



import warnings

from weylchamber import closest_LI, c1c2c3
import qutip


def test_closest_LI_trivial_powell():
    """Test that a gate exactly at (c1, c2, c3) is in fact closest to itself"""
    warnings.filterwarnings(
        'ignore', message='the matrix subclass is not the recommended way')
    CNOT = qutip.gates.cnot()
    U = closest_LI(CNOT, *c1c2c3(CNOT), method='Powell')
    assert isinstance(U, qutip.Qobj)
    assert (CNOT - U).norm() < 1e-15


def test_closest_LI_trivial_leastsq():
    """Test that a gate exactly at (c1, c2, c3) is in fact closest to itself"""
    warnings.filterwarnings(
        'ignore', message='the matrix subclass is not the recommended way')
    CNOT = qutip.gates.cnot()
    U = closest_LI(CNOT, *c1c2c3(CNOT), method='leastsq')
    assert isinstance(U, qutip.Qobj)
    assert (CNOT - U).norm() < 1e-15



"""Top-level package for weylchamber."""

__version__ = '0.3.1'

from .cartan_decomposition import *
from .coordinates import *
from .local_invariants import *
from .perfect_entanglers import *
from .visualize import *
from .gates import *
from . import prec


__all__ = [
    'cartan_decomposition',
    'c1c2c3',
    'canonical_gate',
    'from_magic',
    'point_in_region',
    'point_in_weyl_chamber',
    'random_gate',
    'random_weyl_point',
    'to_magic',
    'weyl_region',
    'g1g2g3',
    'g1g2g3_from_c1c2c3',
    'J_T_LI',
    'closest_LI',
    'project_to_PE',
    'F_PE',
    'concurrence',
    'WeylChamber',
    'bell_basis',
    'gate',
    'mapped_basis',
]



import numpy as np
from numpy import sin, cos, pi
import scipy
from scipy.optimize import leastsq
import logging
import qutip

from .prec import DEFAULT_WEYL_PRECISSION
from ._types import Gate, GTuple
from .coordinates import to_magic, c1c2c3, _SQ_unitary
from .cartan_decomposition import canonical_gate

__all__ = [
    'g1g2g3',
    'g1g2g3_from_c1c2c3',
    'J_T_LI',
    'closest_LI',
    'make_LI_krotov_chi_constructor',
]


def g1g2g3(U: Gate, ndigits=DEFAULT_WEYL_PRECISSION) -> GTuple:
    """Calculate local invariants $(g_1, g_3, g_3)$

    Given a two-qubit gate, calculate local invariants $(g_1, g_2, g_3)$.
    U must be in the canonical basis. For numerical stability, the resulting
    values are rounded to the given precision, cf. the `ndigits` parameter of
    the built-in :func:`round` function.

    >>> print("%.2f %.2f %.2f" % g1g2g3(qutip.gates.cnot()))
    0.00 0.00 1.00
    """
    # mathematically, the determinant of U and UB is the same, but
    # we seem to get better numerical accuracy if we calculate detU with
    # the rotated U
    UB = to_magic(U).full()  # instance of np.ndarray
    detU = np.linalg.det(UB)
    m = UB.T @ UB
    g1_2 = (np.trace(m)) ** 2 / (16.0 * detU)
    g3 = (np.trace(m) ** 2 - np.trace(m @ m)) / (4.0 * detU)
    g1 = round(g1_2.real + 0.0, ndigits)  # adding 0.0 turns -0.0 into +0.0
    g2 = round(g1_2.imag + 0.0, ndigits)
    g3 = round(g3.real + 0.0, ndigits)
    return (g1, g2, g3)


def g1g2g3_from_c1c2c3(
    c1: float, c2: float, c3: float, ndigits=DEFAULT_WEYL_PRECISSION
) -> GTuple:
    """Calculate local invariants from the Weyl chamber coordinates

    Calculate the local invariants $(g_1, g_2, g_3)$ from the Weyl chamber
    coordinates $(c_1, c_2, c_3)$, in units of π. The result is rounded to the
    given precision, in order to enhance numerical stability (cf. `ndigits`
    parameter of the built-in :func:`round` function)

    Example:
        >>> CNOT = qutip.gates.cnot()
        >>> print("%.2f %.2f %.2f" % g1g2g3_from_c1c2c3(*c1c2c3(CNOT)))
        0.00 0.00 1.00
    """
    c1 *= pi
    c2 *= pi
    c3 *= pi
    g1 = round(
        cos(c1) ** 2 * cos(c2) ** 2 * cos(c3) ** 2
        - sin(c1) ** 2 * sin(c2) ** 2 * sin(c3) ** 2
        + 0.0,
        ndigits,
    )
    g2 = round(0.25 * sin(2 * c1) * sin(2 * c2) * sin(2 * c3) + 0.0, ndigits)
    g3 = round(4 * g1 - cos(2 * c1) * cos(2 * c2) * cos(2 * c3) + 0.0, ndigits)
    return g1, g2, g3


def J_T_LI(O: Gate, U: Gate, form='g'):
    """Calculate value of the local-invariants functional

    Args:
        O: The optimal gate
        U: The achieved gate
        form (str): form of the functional to use, 'g' or 'c'
    """
    if form == 'g':
        return np.sum(np.abs(np.array(g1g2g3(O)) - np.array(g1g2g3(U))) ** 2)
    elif form == 'c':
        delta_c = np.array(c1c2c3(O)) - np.array(c1c2c3(U))
        return np.prod(cos(np.pi * (delta_c) / 2.0))
    else:
        raise ValueError("Illegal value for 'form'")


def closest_LI(
    U: Gate, c1: float, c2: float, c3: float, method='leastsq', limit=1.0e-6
):
    """Find the closest gate that has the given Weyl chamber coordinates

    The `c1`, `c2`, `c3` are given in units of π
    """
    A = canonical_gate(c1, c2, c3)

    def f_U(p):
        return _SQ_unitary(*p[:8]) * A * _SQ_unitary(*p[8:])

    return _closest_gate(U, f_U, n=16, method=method, limit=limit)


def _closest_gate(U, f_U, n, x_max=2 * pi, method='leastsq', limit=1.0e-6):
    """Find the closest gate to U that fulfills the parametrization implied by
    the function f_U

    Args:
        U (Gate): Target gate
        f_U (callable): function that takes an array of n values and returns an
            gate.
        n (integer): Number of parameters (size of the argument of f_U)
        x_max (float): Maximum value for each element of the array passed as an
            argument to f_U. There is no way to have different a different
            range for the different elements
        method (str): Name of mimimization method, either 'leastsq' or any of
            the gradient-free methods implemented by scipy.optimize.mimize
        limit (float): absolute error of the distance between the target gate
            and the optimized gate for convergence. The limit is automatically
            increased by an order of magnitude every 100 iterations
    """
    logger = logging.getLogger(__name__)
    logger.debug("_closests_gate with method %s", method)
    from scipy.optimize import minimize

    if method == 'leastsq':

        def f_minimize(p):
            d = _vectorize(f_U(p) - U)
            return np.concatenate([d.real, d.imag])

    else:

        def f_minimize(p):
            return _norm(U - f_U(p))

    dist_min = None
    iter = 0
    while True:
        iter += 1
        if iter > 100:
            iter = 0
            limit *= 10
            logger.debug("_closests_gate limit -> %.2e", limit)
        p0 = x_max * np.random.random(n)
        success = False
        if method == 'leastsq':
            p, info = leastsq(f_minimize, p0)
            U_min = f_U(p)
            if info in [1, 2, 3, 4]:
                success = True
        else:
            res = minimize(f_minimize, p0, method=method)
            U_min = f_U(res.x)
            success = res.success
        if success:
            dist = _norm(U_min - U)
            logger.debug("_closests_gate dist = %.5e", dist)
            if dist_min is None:
                dist_min = dist
                logger.debug("_closests_gate dist_min -> %.5e", dist_min)
            else:
                logger.debug(
                    "_closests_gate delta_dist = %.5e", abs(dist - dist_min)
                )
                if abs(dist - dist_min) < limit:
                    return U_min
                else:
                    if dist < dist_min:
                        dist_min = dist
                        logger.debug(
                            "_closests_gate dist_min -> %.5e", dist_min
                        )


def _vectorize(a, order='F'):
    """Return vectorization of multi-dimensional numpy array or matrix `a`

    Examples:
        >>> a = np.array([1,2,3,4])
        >>> _vectorize(a)
        array([1, 2, 3, 4])

        >>> a = np.array([[1,2],[3,4]])
        >>> _vectorize(a)
        array([1, 3, 2, 4])

        >>> _vectorize(a, order='C')
        array([1, 2, 3, 4])
    """
    if isinstance(a, qutip.Qobj):
        a = a.full()
    N = a.size
    return np.squeeze(np.asarray(a).reshape((1, N), order=order))


def _norm(v):
    """Calculate the norm of a vector or matrix `v`, matching the inner product
    defined in the `inner` routine. An algorithm like
    Gram-Schmidt-Orthonormalization will only work if the choice of norm and
    inner product are compatible.

    If `v` is a vector, the norm is the 2-norm (i.e. the standard Euclidian
    vector norm).

    If `v` is a matrix, the norm is the Hilbert-Schmidt (aka Frobenius) norm.
    Note that the HS norm of a matrix is identical to the 2-norm of any
    vectorization of that matrix (e.g. writing the columns of the matrix
    underneat each other). Also, the HS norm of the m x 1 matrix is the same as
    the 2-norm of the equivalent m-dimensional vector.
    """
    if isinstance(v, qutip.Qobj):
        v = v.data
    if isinstance(v, scipy.sparse.spmatrix):
        return scipy.sparse.linalg.norm(v)
    else:
        return scipy.linalg.norm(v)


def make_LI_krotov_chi_constructor(gate, canonical_basis, unitarity_weight=0):
    r"""Return a constructor for the χ's in an LI optimization.

    Return a `chi_constructor` that determines the boundary condition of the
    backwards propagation in an optimization towards the local equivalence
    class of `gate` in Krotov's method, based on the foward-propagtion of the
    Bell states.

    Args:
        gate (qutip.Qobj): A 4×4 quantum gate, in the `canonical_basis`.
        canonical_basis (list[qutip.Qobj]): A list of four basis states that
            define the canonical basis $\ket{00}$, $\ket{01}$, $\ket{10}$, and
            $\ket{11}$ of the logical subspace.
        unitarity_weight (float): A weight in [0, 1] that determines how much
            emphasis is placed on  maintaining population in the logical
            subspace.

    Returns:
        callable: a function ``chi_constructor(fw_states_T, *args)`` that
        receive the result of a foward propagation of the Bell states (obtained
        from `canonical_basis` via :func:`weylchamber.gates.bell_basis`), and
        returns a list of statex $\ket{\chi}$ that are the boundary condition
        for the backward propagation in Krotov's method. Positional arguments
        beyond `fw_states_T` are ignored.
    """
    # see make_PE_krotov_chi_constructor
    raise NotImplementedError()


def _get_a_kl_PE(UB):
    """Return the 4×4 `A_kl` coefficient matrix (:class:`qutip.Qobj`)
    for the perfect-entanglers functional, for a given gate `UB` in the Bell
    basis.
    """
    raise NotImplementedError()



"""Tests for `weylchamber` package."""

import pytest
from pkg_resources import parse_version

import weylchamber


def test_valid_version():
    """Check that the package defines a valid __version__"""
    assert parse_version(weylchamber.__version__) >= parse_version("0.1.0")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages


def get_version(filename):
    """Extract the package version"""
    with open(filename) as in_fh:
        for line in in_fh:
            if line.startswith('__version__'):
                return line.split('=')[1].strip()[1:-1]
    raise ValueError("Cannot extract version from %s" % filename)


with open('README.rst') as readme_file:
    readme = readme_file.read()

try:
    with open('HISTORY.rst') as history_file:
        history = history_file.read()
except OSError:
    history = ''

requirements = ['numpy', 'scipy', 'matplotlib', 'qutip']

dev_requirements = [
    'coverage', 'pytest', 'pytest-cov', 'pytest-xdist', 'twine', 'pep8',
    'flake8', 'wheel', 'sphinx', 'sphinx-autobuild', 'sphinx_rtd_theme',
    'sphinx-autodoc-typehints', 'gitpython', 'better-apidoc', 'jupyter',
    'nbval', 'nbsphinx', 'watermark', 'click']


version = get_version('./src/weylchamber/__init__.py')

setup(
    author="Michael Goerz",
    author_email='mail@michaelgoerz.net',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    description="Python package for analyzing two-qubit gates in the Weyl chamber",
    install_requires=requirements,
    extras_require={
        'dev': dev_requirements,
    },
    license="BSD license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='weylchamber',
    name='weylchamber',
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    url='https://github.com/qucontrol/weylchamber',
    version=version,
    zip_safe=False,
)



import warnings

import os
import qutip
import numpy as np
from weylchamber import perfect_entanglers


def test_pe_chi_construction_uni(request):
    """Test the co-state construction for the perfect-entanglers functional
    in case of unitary dymanics in the 4x4 subspace"""
    testdir = os.path.splitext(request.module.__file__)[0]
    fw = [0 for i in range(4)]
    chi_exp = [0 for i in range(4)]
    for i in range(4):
        file = os.path.join(testdir, 'fw_{}_uni.dat'.format(i + 1))
        fw[i] = qutip.Qobj(
            np.loadtxt(file, usecols=[0], unpack=True)
            + 1j * np.loadtxt(file, usecols=[1], unpack=True)
        )
        file = os.path.join(testdir, 'chis_{}_uni.dat'.format(i + 1))
        chi_exp[i] = qutip.Qobj(
            np.loadtxt(file, usecols=[0], unpack=True)
            + 1j * np.loadtxt(file, usecols=[1], unpack=True)
        )
        chi_exp[i] = qutip.Qobj(chi_exp[i])
    psi_00 = qutip.Qobj(np.array([1, 0, 0, 0]))
    psi_01 = qutip.Qobj(np.array([0, 1, 0, 0]))
    psi_10 = qutip.Qobj(np.array([0, 0, 1, 0]))
    psi_11 = qutip.Qobj(np.array([0, 0, 0, 1]))
    chi_constructor = perfect_entanglers.make_PE_krotov_chi_constructor(
        [psi_00, psi_01, psi_10, psi_11]
    )
    chi_out = chi_constructor(fw)
    for i in range(4):
        assert abs((chi_out[i] - chi_exp[i]).norm()) < 1e-12


def test_pe_chi_construction_nonuni(request):
    """Test the co-state construction for the perfect-entanglers functional
    in case of non-unitary dymanics in the 4x4 subspace"""
    testdir = os.path.splitext(request.module.__file__)[0]
    fw = [0 for i in range(4)]
    chi_exp = [0 for i in range(4)]
    for i in range(4):
        file = os.path.join(testdir, 'fw_{}_nonuni.dat'.format(i + 1))
        fw[i] = qutip.Qobj(
            np.loadtxt(file, usecols=[0], unpack=True)
            + 1j * np.loadtxt(file, usecols=[1], unpack=True)
        )
        file = os.path.join(testdir, 'chis_{}_nonuni.dat'.format(i + 1))
        chi_exp[i] = qutip.Qobj(
            np.loadtxt(file, usecols=[0], unpack=True)
            + 1j * np.loadtxt(file, usecols=[1], unpack=True)
        )
        chi_exp[i] = qutip.Qobj(chi_exp[i])
    psi_00 = qutip.Qobj(np.array([1, 0, 0, 0, 0]))
    psi_01 = qutip.Qobj(np.array([0, 1, 0, 0, 0]))
    psi_10 = qutip.Qobj(np.array([0, 0, 1, 0, 0]))
    psi_11 = qutip.Qobj(np.array([0, 0, 0, 1, 0]))
    chi_constructor = perfect_entanglers.make_PE_krotov_chi_constructor(
        [psi_00, psi_01, psi_10, psi_11], unitarity_weight=0.1
    )
    chi_out = chi_constructor(fw)
    for i in range(4):
        assert abs((chi_out[i] - chi_exp[i]).norm()) < 1e-12
